# xcavation v0.4.0
# By Hunter Brooks, at UToledo, Toledo: Jan. 28, 2026
from .aperture import *
from .genspec import *
from .motion import *
from .quality import *

__version__ = '0.4.0'
__author__ = 'Hunter Brooks'
__email__ = 'hbrooks8@rockets.utoledo.edu'




#-----------------------------------------------------------------------#
# xcavation.genspec v0.4.0
# By Hunter Brooks, at UToledo, Toledo: Jan. 28, 2026
#
# Purpose: Main API Function for SphereX Data Retrieval and Photometry
#-----------------------------------------------------------------------#



# Import Internal Modules
# ------------------------------------------------------ #
from .aperture import spherex_aperature_phot
from .motion import decimal_year, time_mjd, proper_motion
from .quality import finder_chart, spectra_plot
# ------------------------------------------------------ #


# Import Data Management
# ------------------------------------------------------ #
import time
import numpy as np
import pandas as pd
import astropy.units as u
from astropy.time import Time
from datetime import datetime
from astropy.table import Table
from astroquery.ipac.irsa import Irsa
from dataclasses import dataclass, field
# ------------------------------------------------------ #



# Import WCS, Photometry, and Plotting
# ------------------------------------------------------ #
from astropy.coordinates import SkyCoord
# ------------------------------------------------------ #



# Import Multithreading and Watching Packages
# ------------------------------------------------------ #
from tqdm import tqdm
from concurrent.futures import ThreadPoolExecutor
# ------------------------------------------------------ #



# User Configuration Data
# ------------------------------------------------------ #
def variable_verify(ra, dec, # Core Inputs
                    r_fwhm, r_annulus_in, r_annulus_out, # Photometry Radii
                    pmra, pmdec, mjd, # Proper Motion Propagation
                    save_data, output_path, # Saved Data
                    threads, # Number of Multi-Threads
                    enable_print, ram_download, retry_count, # User Control
                    clean_type, bad_bits, # Cleaning Flux Data
                    background_type):
  """
  Verify user input parameters for SPHEREx aperture photometry.

  Parameters
  ----------
    ra: R.A. of the target in degrees (0 <= ra <= 360) (float)
    dec: Decl. of the target in degrees (-90 <= dec <= 90) (float)
    r_fwhm: Aperture radius in units of PSF FWHM (float)
    r_annulus_in: Inner radius of background annulus (float)
    r_annulus_out: Outer radius of background annulus (float)
    pmra: Proper motion in R.A. (float)
    pmdec: Proper motion in Decl. (float)
    mjd: MJD corresponding to the input coordinates (float)
    save_data: Whether to save the photometry output (bool)
    output_path: Path for saving the output file (str)
    threads: Number of CPU threads to allocate (int)
    enable_print: Enable or disable console print statements (bool)
    ram_download: If True, download FITS entirely into memory (bool)
    retry_count: Number of times to retry downloading a file (int)
    clean_type: Method for cleaning flux data (str)
    bad_bits: List of bit indices in FLAG extension (list)

  Returns
  -------
    bool
        True if all input variables are valid, False otherwise
  """

  # Checks R.A. Decl. Inputs
  if ((type(ra) is not float) or (type(dec) is not float)
                or (0 >= ra >= 360) or (-90 >= dec >= 90)):
    print('Please Input Valid: R.A. (deg, float) or Decl. (deg, float)')
    return False

  # Checks PMRA, PMDEC, and MJD Inputs
  if (((type(pmra) is not float) and (type(pmra) is not int))
      or ((type(pmdec) is not float) and (type(pmdec) is not int))
               or (type(mjd) is not float and type(mjd) is not int)):
    print('Please Input Valid: PMRA (float), PMDEC (float), or MJD (float)')
    print('Read Documentation: https://github.com/huntbrooks85/Xcavation')
    return False

  # Checks Aperture Radius Input
  if (type(r_fwhm) is not float and type(r_fwhm) is not int) or ((r_fwhm) < 0):
    print('Please Input Valid: Aperture Radius')
    print('Read Documentation: https://github.com/huntbrooks85/Xcavation')
    return False

  # Checks Inner Annulus Radius Input
  if ((type(r_annulus_in) is not float and type(r_annulus_in) is not int)
                                                  or ((r_annulus_in) < 0)):
    print('Please Input Valid: Inner Annulus Radius')
    print('Read Documentation: https://github.com/huntbrooks85/Xcavation')
    return False

  # Checks Outer Annulus Radius Input
  if ((type(r_annulus_out) is not float and type(r_annulus_out) is not int)
                                                    or ((r_annulus_out) < 0)):
    print('Please Input Valid: Outer Annulus Radius')
    print('Read Documentation: https://github.com/huntbrooks85/Xcavation')
    return False

  # Checks Saved Data Inputted Variable
  if type(save_data) is not bool:
    print('Please Input Valid: Save_data (True or False)')
    print('Read Documentation: https://github.com/huntbrooks85/Xcavation')
    return False

  # Checks Output Path Inputted
  if (type(output_path) is not str):
    print('Please Input Valid: Spectrum Table (string)')
    print('Read Documentation: https://github.com/huntbrooks85/Xcavation')
    return False

  # Checks Inputted Threads
  if (type(threads) is not int) or (1 >= threads >= 50):
    print('Please Input a Valid: Amount of Allocated CPU Threads')
    print('Read Documentation: https://github.com/huntbrooks85/Xcavation')
    return False

  # Checks Inputted Print Variables
  if (type(enable_print) is not bool):
    print('Please Input a Valid: Enable Print (True or False)')
    print('Read Documentation: https://github.com/huntbrooks85/Xcavation')
    return False

  # Checks RAM Download Variable
  if (type(ram_download) is not bool):
    print('Please Input a Valid: Ram Download (True or False)')
    print('Read Documentation: https://github.com/huntbrooks85/Xcavation')
    return False

  # Checks Download Retry Count Variables
  if (type(retry_count) is not int) or (retry_count < 1):
    print('Please Input a Valid: Retry Count (integer)')
    print('Read Documentation: https://github.com/huntbrooks85/Xcavation')
    return False

  # Checks Clean Type Variable
  if ((clean_type != 'none') and (clean_type != 'lacosmic')
      and (clean_type != 'mask') and (clean_type != 'median_mask')
        and (clean_type != 'interp_mask')):
    print('Please Input a Valid: Clean Tyle (none, laplacian, mask)')
    print('Read Documentation: https://github.com/huntbrooks85/Xcavation')
    return False

  # Checks Background Type
  if ((background_type != 'mean') and (background_type != 'median')
      and (background_type != 'mode')):
    print('Please Input a Valid: Background Type (mean, median, mode)')
    print('Read Documentation: https://github.com/huntbrooks85/Xcavation')
    return False

  # Checks Flagged Variable List
  if (type(bad_bits) is not list) or (len(bad_bits) < 1):
    print('Please Input a Valid: Bad Bits (list)')
    print('Read Documentation: https://github.com/huntbrooks85/Xcavation')
    return False

  # Returns True if Every Variable is Good
  return True
# ------------------------------------------------------ #



# User Configuration Data
# ------------------------------------------------------ #
@dataclass
class genspec_profile:
    r_fwhm: float = 2 # Aperture Radius FWHM
    r_annulus_in: float = 7.5 # Inner Annulus Radius FWHM
    r_annulus_out: float = 15 # Outer Annulus Radius FWHM
    pmra: float = 0 # Proper Motion R.A. (arcsec/yr)
    pmdec: float = 0 # Proper Motion Decl. (arcsec/yr)
    mjd: float = 61000 # Modified Julian Date
    save_data: bool = False # Save Q.A., Table, and Spectral Data
    output_path: str = 'xcavation_save' # Saved Location
    threads: int = 1 # Number of Threads for Multi-Threading
    enable_print: bool = True # Whether to Print
    ram_download: bool = False # Download Images to RAM
    retry_count: int = 10 # How Many HTML Retries
    clean_type: str = 'none' # What type of Image Cleaning
    bad_bits: list = field(default_factory=lambda: [0,1,10,11]) # Flags Removed
    background_type: str = 'mean' # What type of background subtraction
# ------------------------------------------------------ #



# Creates Multi-Try Query
# ------------------------------------------------------ #
def retry(func, *args, retries=10, delay=0.5, **kwargs):
    """
    Execute a function with automatic retries on failure.

    Parameters
    ----------
      func: Called function
      *args: All relavent data for called function
      retries : Max number of retries (int)
      delay: Delay between retries (float)

    Returns
    -------
      result: object or None
          Return value of `func` if successful; otherwise None if all
          retries fail.
    """



    # ----- Repeat Calling Loop ----- #
    for attempt in range(1, retries + 1):
        try: # Call main function (aperture photometry)
            return func(*args, **kwargs)
        except Exception as e: # If failed retry (usual fail is HTML)
            if attempt < retries: # Keep Going if more retries
                time.sleep(delay)
            else: # Fail if max retries reached
                return None
    # ------------------------------- #
# ------------------------------------------------------ #



# Multi-Thread the SphereX Query Function
# ------------------------------------------------------ #
def genspec(ra, dec, config: genspec_profile):
    """
    Generate a SPHEREx spectrophotometric spectrum at a given sky position.
    This is the main function for the xcavation SPHEREx
    spectrophotometry tool. It performs multi-threaded SPHEREx aperture
    photometry calculations while optionally produces finder charts, spectral
    plots, and spectral table.

    Parameters
    ----------
      ra: R.A. of the target in degrees (float).
      dec: Decl. of the target in degrees (float).
      config: User generated configuration for queried data

    Returns
    -------
      output : pandas.DataFrame
          - wavelength: float
              Wavelength at queried location
          - delta_lambda: float
              Spectral resolution using R value solution
          - flux: float
              Aperture flux in μJy
          - flux_err: float
              Propagated flux uncertainty in μJy
          - flag_count: str
              Dictionary of flagged pixel counts per bit
          - SNR: float
              Signal-to-noise ratio of given fluxes
          - flux_cutout: array
              Flux cutout
          - flag_cutout: array
              FLAG cutout
          - aperture: array
              Boolean aperture cutout
          - annulus: array
              Boolean annulus cutout
          - x_loc, y_loc: float
              Pixel coordinates of the target in the cutout
          - ap_radius: float
              Aperture radius
          - inner_annulus, outer_annulus : float
              Annulus radii
          - url: str
              Source FITS URL.
          - mjd: float
              Observation MJD

    Raises
    ------
      ValueError
          If input variables fail validation or network failure
    """



    # ----- Obtain User Config. Data ----- #
    r_fwhm = config.r_fwhm # Radius in FWHM
    r_annulus_in = config.r_annulus_in # Inner Annulus Radius in FWHM
    r_annulus_out = config.r_annulus_out # Outer Annulus Radius in FWHM
    pmra = config.pmra # Proper Motion of R.A. in arcsec/yr
    pmdec = config.pmdec # Proper Motion of Decl. in arcsec/yr
    mjd = config.mjd # R.A. and Decl. Start MJD
    save_data = config.save_data # Boolean on Whether to Save Data
    output_path = config.output_path # Path to Save Data
    threads = config.threads # Number of Multi-Threads
    enable_print = config.enable_print # Boolean on Whether to Print Results
    ram_download = config.ram_download # Download FITS Images to RAM
    bad_bits = config.bad_bits # Which Flags to Mask
    retry_count = config.retry_count # Number of Retries
    clean_type = config.clean_type # Which Method of Bad Pixel Fixes to USE
    background_type = config.background_type # Which type of background subtraction
    # ------------------------------------ #



    # ----- Verify User Inputted Variables ----- #
    verify = variable_verify(ra, dec, # Positional Data
                             r_fwhm, r_annulus_in, r_annulus_out, # Radii Data
                             pmra, pmdec, mjd, # Proper Motion Data
                             save_data, output_path, # Saved Data
                             threads, enable_print, ram_download, retry_count, # Misc.
                             clean_type, bad_bits, # Masking Data
                             background_type) # Background Type

    # Raise Error of User Inputted Variables is Bad
    if verify is False:
      raise ValueError('Poorly Assigned Variable. Please Read Documentation.')
    # ------------------------------------------ #



    # ----- Proper Motion Propagation ----- #
    # Calculate Propagated Proper Motion
    mjd_now = Time(datetime.now()).mjd # Current MJD
    time_passed = time_mjd(mjd, mjd_now) # Years since observation
    ra_deg, dec_deg = proper_motion(ra, dec,
                          pmra, pmdec, time_passed) # Adjust for Proper Motion

    # Print Current R.A. and Decl. Data if Enable Print is True
    if enable_print is True:
      print('\n+====================================================================+')
      print('       Starting xcavation.genspec SPHEREx Spectrophotometry Tool')
      print('+====================================================================+')
      print(f'''Queried Coordinates: R.A. = {round(ra_deg, 6)} (deg), Decl. = {round(dec_deg, 6)} (deg)''') # Print Adjusted Coordinates
      print(f"Query Time: {datetime.now()}")
      print('+====================================================================+\n')
    # ------------------------------------- #



    # ----- SphereX Query ----- #
    # Query For All SPHEREx URLs in a Nearby Search
    coord = SkyCoord(ra_deg, dec_deg, unit='deg') # Convert to astropy Units
    results = Irsa.query_sia(pos=(coord, 10 * u.arcsec),
              collection='spherex_qr2') # Query SphereX QR2 (Will Need Updates)
    urls = results["access_url"] # Gets the Image URLs

    # Read in the Number of Each Detectors
    D1, D2, D3, D4, D5, D6 = 0, 0, 0, 0, 0, 0 # Creates Empty Numbers
    for url in urls:
      # Splits URL to Obtain Detector Number
      num_detector = int(url.split('D')[len(url.split('D')) - 1][0])
      if num_detector == 1: # Detector 1
        D1 += 1
      if num_detector == 2: # Detector 2
        D2 += 1
      if num_detector == 3: # Detector 3
        D3 += 1
      if num_detector == 4: # Detector 4
        D4 += 1
      if num_detector == 5: # Detector 5
        D5 += 1
      if num_detector == 6: # Detector 6
        D6 += 1
    # ------------------------- #



    # ----- Multi-Thread Wavelength and Aperature Calc. ----- #
    output = [] # Creates Empty List
    with ThreadPoolExecutor(max_workers=threads) as executor:

        # Multi-Thread w/ tqdm
        futures = [executor.submit(retry, spherex_aperature_phot, url,
                          [ra, dec], [pmra, pmdec], mjd,
                            r_fwhm, r_annulus_in, r_annulus_out,
                              ram_download, clean_type, bad_bits,
                                   background_type,
                                   retries = retry_count) for url in urls]

        # Prints Start of Query
        if enable_print is True:
          print('+====================================================================+')
          print(f'{len(urls)} Images Were Found in SPHEREx')

        # Sets Up Progress Bar
        for future in tqdm(futures, desc="Calculating", ncols=70, colour = '#C5D89D',
                           bar_format='''{l_bar}{bar}|{n_fmt}/{total_fmt} [{elapsed}<{remaining}]''',
                           ascii="░▒█", disable=not enable_print):

            # Starts Multi-Thread
            result = future.result()

            # Raises Error if Network is Down
            if result is None:
              raise ValueError('Network Connectivity Issue. Please Try Again')

            # Adds to Empty Array For Total Dictionary
            output.append({
                "wavelength": result["wavelength"], # Wavelength
                "delta_lambda": result["delta_lambda"], # Resolving Power
                "flux": result["flux"], # Aperture Flux
                "flux_err": result["flux_err"], # Aperture Flux Error
                "flag_count": str(result["flag_count"]), # Dict. of Flags
                "SNR": result["SNR"], # Signal-to-Noise Ratio
                "flux_cutout": result["flux_cutout"], # Flux Cutout
                "flag_cutout": result["flag_cutout"], # Bool. Flag Cutout
                "aperture": result["aperture"], # Aperture Mask
                "annulus": result["annulus"], # Annulus Mask
                "x_loc": result["x_loc"], # X Pixel Location in Cutout
                "y_loc": result["y_loc"], # Y Pixel Location in Cutout
                "ap_radius": result['ap_radius'], # Aperture Radius
                "inner_annulus": result['inner_annulus'], # Inner Annulus Radius
                "outer_annulus": result['outer_annulus'], # Outer Annulus Radius
                "url": result["url"], # Image URL
                "mjd": result["mjd"] # Observation MJD
            })

        # Prints Finished Query
        if enable_print is True:
          print('Spectral Calulcation Complete')
          print('+====================================================================+\n')

    # Converts Final Dictionary to pandas DataFrame
    output = pd.DataFrame(output)
    # ------------------------------------------------------- #



    # ----- Finder Chart and Spectral Plots ----- #
    # Only Queries if Selected
    if save_data is True:

      # Start Q.A. Finder Chart Creation
      finder_chart(output, output_path + '.pdf')

      # Prints Finished Finder Chart
      if enable_print is True:
        print('+====================================================================+')
        print('Finder Chart Was Successfully Saved')
        print(f'Spectral Plot of: R.A. = {round(ra, 3)} (deg) and Decl. = {round(dec, 3)} (deg)')

      # Plots Spectral Plot
      spectra_plot(output)

      # Gets Relevant Data for Table
      lambda_list = np.array(output['wavelength']) # Wavelength
      flux_list = np.array(output['flux']) # Flux
      flux_err_list = np.array(output['flux_err']) # Flux Error
      delta_lambda = np.array(output['delta_lambda']) # Resolving Power
      flag_count = np.array(output['flag_count']) # Flag Dictionary

      # Sort Relevant Data Based on Wavelength
      sort_idx = np.argsort(lambda_list) # Sort Based on Wavelength
      lambda_list = lambda_list[sort_idx] # Sort Wavelength
      flux_list = flux_list[sort_idx] # Sort Flux
      flux_err_list = flux_err_list[sort_idx] # Sort Flux Error
      delta_lambda = delta_lambda[sort_idx] # Sort Resolving Power
      flag_count = flag_count[sort_idx] # Sort Flag Dictionary

      # Creates Table with Header
      t = Table([lambda_list, flux_list, flux_err_list,
              delta_lambda, flag_count],
          names=('lambda', 'flux', 'flux_err', 'delta_lambda', 'flag_count'))

      # Opens Created Files To Edit
      with open(output_path + '.txt', 'w') as f:

          # Generates Relevant Table Header
          f.write(f'# Description: Generated Using xcavation.genspec SPHEREx Spectrophotometry Tool\n')
          f.write(f'# Publisher: IPAC/IRSA\n')
          f.write(f'# Contact: Mr. Hunter Brooks\n')
          f.write(f'# Email: hbrooks8@rockets.utoledo.edu\n')
          f.write(f'# Data Source: SPHEREx QR2 Images via IRSA\n')
          f.write(f'# Title: Spectrum (ASCII)\n')
          f.write(f'# RA: {ra}\n')
          f.write(f'# Dec: {dec}\n')
          f.write(f'# Wavelength Range: {round(np.nanmin(output['mjd']), 3)} to {round(np.nanmax(output['mjd']), 3)}\n')
          f.write(f'# Spectral resolution (Δλ/λ): 35 – 130\n')
          f.write(f'# Number of images measured: {len(urls)}\n')
          f.write(f'# Photometry Type: Aperture\n')
          f.write(f'# Average SNR: {round(np.nanmean(output['flux']/output['flux_err']), 2)}\n')
          f.write(f'# Average Flux Error: {round(np.nanmean(output['flux_err']))}\n')
          f.write(f'# Images per detector: D1={D1}, D2={D2}, D3={D3}, D4={D4}, D5={D5}, D6={D6}\n')
          f.write(f'# References: https://spherex.caltech.edu/page/publications; https://github.com/huntbrooks85/Xcavation\n')
          f.write(f'# Column 1: Wavelength (micron)\n')
          f.write(f'# Column 2: Flux (uJy)\n')
          f.write(f'# Column 3: Flux Error (uJy)\n')
          f.write(f'# Column 4: Delta Lambda (microns)\n')
          f.write(f'# Column 5: Aperture Flag Bit Count (dictionary)\n')
          t.write(f, format='ascii.commented_header', overwrite=True)

      # Print Finished Saving Data
      if enable_print is True:
        print('Spectrum Table Was Successfully Saved')
        print('+====================================================================+\n')
    # ------------------------------------------- #



    # ----- Print xcavation Summary ----- #
    if enable_print is True:
      print('+====================================================================+')
      print('              xcavation Spectrophotometry Tool Summary')
      print('+====================================================================+')

      # Print Coordinate Data
      print(f"Queried Coordinates: R.A. = {round(ra_deg, 6)} (deg), Decl. = {round(dec_deg, 6)} (deg)") # Print Adjusted Coordinates
      print(f"SIMBAD Query: https://simbad.u-strasbg.fr/simbad/sim-coo?Coord={ra}+{dec}&CooFrame=FK5&CooEpoch=2000&CooEqui=2000&CooDefinedFrames=none&Radius=0.5&Radius.unit=arcmin&submit=submit+query&CoordList=")
      print(f"Query Finished: {datetime.now()}\n")

      # Print Number of Image Data
      print(f"Number of Images Found: {len(urls)}")
      print(f"Images Per Detector: D1={D1}, D2={D2}, D3={D3}, D4={D4}, D5={D5}, D6={D6}\n")

      # Print Flux Data
      print(f'Average SNR: {round(np.nanmean(output['flux']/output['flux_err']), 2)}')
      print(f'Average Flux Error: {round(np.nanmean(output['flux_err']))}\n')

      # Print Spectral Data
      print(f'Observation Range: {round(np.nanmin(output['mjd']), 1)} to {round(np.nanmax(output['mjd']), 1)}')
      print(f'Wavelength Range: {round(np.nanmin(output['wavelength']), 3)} to {round(np.nanmax(output['wavelength']), 3)}')
      print(f"Flux Range: {round(np.nanmin(output['flux']), 3)} to {round(np.nanmax(output['flux']), 3)}")

      # Print Where Data Was Saved
      if save_data == True:
        print(f'\nData Saved To: {output_path}')
      print('+====================================================================+')
    # ----------------------------------- #



    # Returns The Output Dictionary
    return output
# ------------------------------------------------------ #




#-----------------------------------------------------------------------#
# xcavation.quality v0.4.0
# By Hunter Brooks, at UToledo, Toledo: Jan. 28, 2026
#
# Purpose: Plotting Tools for SphereX Data
#-----------------------------------------------------------------------#



# Import Data Management
# ------------------------------------------------------ #
import math
import numpy as np
# ------------------------------------------------------ #



# Import WCS, Photometry, and Plotting
# ------------------------------------------------------ #
import matplotlib.pyplot as plt
from matplotlib.patches import Circle
from matplotlib.colors import ListedColormap
# ------------------------------------------------------ #



# Plot Style
# ------------------------------------------------------ #
plt.style.use('seaborn-v0_8-paper')
plt.rcParams.update({
  'font.family': 'STIXGeneral',
  'font.size': 14,
  'axes.labelsize': 20,
  'axes.titlesize': 25,
  'xtick.labelsize': 15,
  'ytick.labelsize': 15,
  'axes.linewidth': 1.2,
  'legend.fontsize': 12
                  })
  # ------------------------------------------------------ #




# Function to Generate Q.A. Finder Charts
# ------------------------------------------------------ #
def finder_chart(output, pdf_path):
  """
  Generate a multi-panel finder chart PDF from SPHEREx aperture photometry results.

  This function creates a grid of cutout images for multiple wavelengths,
  overlays the aperture and annulus regions, and marks the target position.
  The panels are sorted by wavelength. The final figure is saved as a PDF.

  Parameters
  ----------
    output: dict
        Dictionary containing SPHEREx photometry:
        - 'flux_cutout': list of 2D flux arrays (μJy)
        - 'aperture': list of 2D aperture masks
        - 'annulus': list of 2D annulus masks
        - 'x_loc', 'y_loc': lists of pixel coordinates of the target in cutouts
        - 'wavelength': list or array of wavelengths (μm)
        - 'flux': list of aperture flux values (μJy)
        - 'ap_radius', 'inner_annulus', 'outer_annulus': lists of radii in arcsec
    pdf_path: File path where the PDF finder chart will be saved (str)

  Returns
  -------
    None: Figure save to the specified path
  """



  # ------- Set-Up Subplots ------- #
  n_images = len(output['flux_cutout']) # Total number of flux arrays
  ncols = 5 # Number of Columns (pre-set)
  nrows = math.ceil(n_images / ncols) # Number of Rows (based on total)
  # ------------------------------- #



  # ------- Get Required Data From Input Array ------- #
  # Sort By Wavelength
  wavelengths = np.array(output['wavelength']) # Original Wavelength List
  sorted_idx = np.argsort(wavelengths) # Sorted Wavelength List

  # Sorts All Needed Variables
  flux_cutouts_sorted = [output['flux_cutout'][i] for i in sorted_idx] # Cutouts
  apers_sorted = [output['aperture'][i] for i in sorted_idx] # Aperture Mask
  annuls_sorted = [output['annulus'][i] for i in sorted_idx] # Annulus Mask
  x_locs_sorted = [output['x_loc'][i] for i in sorted_idx] # X Pixel Location
  y_locs_sorted = [output['y_loc'][i] for i in sorted_idx] # Y Pixel Location
  wavelengths_sorted = wavelengths[sorted_idx] # Sorted Wavelength
  fluxes_sorted = [output['flux'][i] for i in sorted_idx] # Flux
  aperture_sorted = [output['ap_radius'][i] for i in sorted_idx] # Aperture Radius
  inner_annulus_sorted = [output['inner_annulus'][i] for i in sorted_idx] # Inner Annulus Radius
  outer_annulus_sorted = [output['outer_annulus'][i] for i in sorted_idx] # Outer Annulus Radius
  # -------------------------------------------------- #



  # ------- Create Subplot ------- #
  fig, axes = plt.subplots(nrows, ncols, figsize=(3*ncols, 3*nrows))
  # ------------------------------ #



  # ------------ Plot Each Cutout ------------ #
  axes = axes.ravel()
  valid_count = 0
  for i in range(n_images):
      # Get Relavent Data
      cutouts = flux_cutouts_sorted[i] # Cutout
      wavelength = wavelengths_sorted[i] # Wavelength
      flux = fluxes_sorted[i] # Flux

      # Verify that each Cutout, Wavelength, and Flux is Valid
      if (cutouts is not None) and (np.isfinite(wavelength)) and (np.isfinite(flux)):
        ax = axes[valid_count] # Counts Valid Axes

        # Get the Valid Data
        apers = apers_sorted[i] # Aperture Mask
        annuls = annuls_sorted[i] # Annulus Mask
        x_loc = x_locs_sorted[i] # Queried X Location
        y_loc = y_locs_sorted[i] # Queried Y Location
        aperture = aperture_sorted[i] # Aperture Radius
        inner_annulus = inner_annulus_sorted[i] # Inner Annulus Radius
        outer_annulus = outer_annulus_sorted[i] # Outer Annulus Radius

        # Plots the Image Data
        ax.scatter(x_loc, y_loc, c='yellow',marker='*', s = 100) # Queried Point
        ax.imshow(cutouts, cmap='Greys',
          vmin=np.nanpercentile(cutouts, 1), vmax=np.nanpercentile(cutouts, 99))
        ax.imshow(np.ma.masked_where(apers == 0, apers),
                  cmap=ListedColormap(['red']), alpha=0.2) # Aperture Mask
        ax.imshow(np.ma.masked_where(annuls == 0, annuls),
                  cmap=ListedColormap(['blue']), alpha=0.2) # Annulus Mask

        # Creates Matplotlib Circles for Aperture and Annulus Radii
        ap = Circle((x_loc, y_loc), aperture/6.15,
                    color='red', lw = 2, fill = False) # Aperture Radius
        in_an = Circle((x_loc, y_loc), inner_annulus/6.15,
                    color='blue', lw = 2, fill = False) # Inner Annulus Radius
        out_an = Circle((x_loc, y_loc), outer_annulus/6.15,
                    color='blue', lw = 2, fill = False) # Outer Annulus Radius

        # Plots Each Circle
        ax.add_patch(ap) # Aperture
        ax.add_patch(in_an) # Inner Annulus
        ax.add_patch(out_an) # Outer Annulus

        # General Subplot Looks
        ax.set_title(f"$λ$: {wavelength:.4f} μm and $F_λ$: {flux:.0f}", fontsize=12)
        ax.axis('off') # Turn off axis
        valid_count += 1 # Counts Valid Plots
  # ------------------------------------------ #



  # ------------ Fix Plots ------------ #
  # Remove Unused Plots
  for ax in axes[valid_count:]:
      fig.delaxes(ax)

  # Adjust Each Subplot
  plt.subplots_adjust(left=0.02, right=0.98,
                      bottom=0.02, top=0.95,
                      wspace=0.05, hspace=0.15)
  # -------------------------------------- #




  # ------------ Save Figure ------------ #
  plt.savefig(pdf_path, bbox_inches='tight') # Save Figure
  plt.close() # Close Figure to Save Memory
  # ------------------------------------- #
# ------------------------------------------------------ #



# Plots Spectra Without Saving
# ------------------------------------------------------ #
def spectra_plot(output):
  """
  Plot a SPHEREx spectrum with flux uncertainties.

  Parameters
  ----------
    output : dict
        Dictionary containing SPHEREx photometry results. Expected keys:
        - 'wavelength': list or array of wavelengths in microns (μm)
        - 'flux': list or array of flux values in μJy
        - 'flux_err': list or array of flux uncertainties in μJy

  Returns
  -------
    None
        Displays the spectrum plot using matplotlib. Does not return data.
  """



  # ------- Create Plot ------- #
  plt.figure(figsize=(6,4))
  # --------------------------- #



  # ------- Get Relevant Data ------- #
  wavelengths = np.array(output['wavelength']) # Wavelength
  fluxes = np.array(output['flux']) # Flux
  errors = np.array(output['flux_err']) # Flux Error
  # --------------------------------- #



  # ------- Sort Data Based On Wavelength ------- #
  sort_idx = np.argsort(wavelengths) # Sort Based on Wavelength
  wavelengths = wavelengths[sort_idx] # Wavelength Sort
  fluxes = fluxes[sort_idx] # Flux Sort
  errors = errors[sort_idx] # Flux Error Sort
  finite = np.isfinite(fluxes) # Ensures that all Flus Data is Real
  # --------------------------------------------- #



  # ------- Plot Flux and Errors ------- #
  plt.errorbar(wavelengths[finite], fluxes[finite], # Plot Flux
               yerr=errors[finite], fmt='o', color='slategray',
               ecolor='slategray', elinewidth=1.5, capsize=3,
               markersize=1, alpha = 0.5)
  plt.step(wavelengths[finite], fluxes[finite], # Plot Errors
           color='#656D3F', alpha=0.6, where = 'mid')
  # ------------------------------------ #



  # ------- Make Plot Pretty ------- #
  plt.xlabel("Wavelength [μm]", fontsize=14) # X Label
  plt.ylabel("Flux [μJy]", fontsize=14) # Y Label

  plt.grid(alpha=0.2) # Plot Grid
  plt.minorticks_on() # Plot Minor Tickmarks
  plt.tight_layout() # Forces Plot to Be Tight on all Paddings

  # Make Y Limit Using Percentiles
  bot = np.nanpercentile(fluxes[finite], 1) - 100
  top =  np.nanpercentile(fluxes[finite], 95) + 100
  plt.ylim(bot, top)

  # Show Plot
  plt.show()
  # -------------------------------- #
# ------------------------------------------------------ #




#-----------------------------------------------------------------------#
# xcavation.motion v0.4.0
# By Hunter Brooks, at UToledo, Toledo: Jan. 28, 2026
#
# Purpose: Propogate celestial coordinates using proper motion
#-----------------------------------------------------------------------#



# Import Data Management
# ------------------------------------------------------ #
import numpy as np
from astropy.time import Time
from datetime import datetime
# ------------------------------------------------------ #




# Calculate Decimal Year
#-----------------------------------------------------------------------#
def decimal_year(t):
    """
    Convert a calendar date to a decimal year.

    Parameters
    ----------
      t: Calendar date and time to be converted.

    Returns
    -------
      decimal_year:  Date expressed as a decimal year (float).
    """



    # ----- Get Years ----- #
    # The Years Closest to Inputted Years
    year_start = datetime(t.year, 1, 1) # The Start of the Inputted year
    next_year_start = datetime(t.year + 1, 1, 1) # The Start of the Next Year
    # --------------------- #



    # ----- Get Difference ----- #
    # The Length of Years From Above
    year_length = (next_year_start - year_start).total_seconds()
    seconds_into_year = (t - year_start).total_seconds()
    # -------------------------- #



    return t.year + seconds_into_year / year_length
#-----------------------------------------------------------------------#



# Calculate Time Passed Since Inputted MJD
#-----------------------------------------------------------------------#
def time_mjd(mjd, mjd_obs):
    """
    Compute the elapsed time in years between two epochs given in
    Modified Julian Date (MJD).

    Parameters
    ----------
      mjd: Reference MJD corresponding to the inputted epoch (float)
      mjd_obs: Observed MJD corresponding to the observed epoch (float)

    Returns
    -------
      time_passed: Time elapsed from `mjd` to `mjd_obs` in decimal years (float)
    """



    # ----- MJD Input ----- #
    # Convert MJD (input) to decimal year (i.e. 2026.2315)
    t = Time(mjd, format='mjd')
    decimal_year_inputted = t.decimalyear
    # --------------------- #



    # ----- MJD Observed ----- #
    # Convert MJD (observed) to decimal year (i.e. 2026.2315)
    t = Time(mjd_obs, format='mjd')
    decimal_year_obs = t.decimalyear
    # ------------------------ #



    # Get the Time of Final Time Propagated
    time_passed = decimal_year_obs - decimal_year_inputted # Calc. Time Passed
    return time_passed
#-----------------------------------------------------------------------#



# Adjust RA and Dec for Proper Motion
#-----------------------------------------------------------------------#
def proper_motion(ra, dec, pmra, pmdec, time_passed):
    """
    Propagate right ascension and declination using proper motion.

    Parameters
    ----------
      ra: Initial R.A. in degrees (float)
      dec : Initial Decl. in degrees (float)
      pmra: Proper motion in R.A. in arcseconds per year (float)
      pmdec: Proper motion in Decl. in arcseconds per year (float)
      time_passed: Time interval over which to propagate the position (float)

    Returns
    -------
      ra_deg: Propagated R.A. in degrees (float)
      dec_deg: Propagated Decl. in degrees (float)
    """



    # ----- Delta R.A. and Decl. ----- #
    dRA  = (pmra  * time_passed) / (3600 * np.cos(np.deg2rad(dec))) # Delta R.A.
    dDec = (pmdec * time_passed) / 3600 # Delta Decl.
    # -------------------------------- #



    # ----- Final R.A. and Decl. ----- #
    ra_deg  = ra  + dRA # R.A.
    dec_deg = dec + dDec # Decl.
    # -------------------------------- #



    return ra_deg, dec_deg
#-----------------------------------------------------------------------#



# setup.py
from setuptools import setup, find_packages

setup(
    name="xcavation",
    version="0.4.0",
    license="MIT License",
    packages=find_packages(),
    install_requires=[
            "lacosmic",
            "numpy",
            "scipy",
            "pandas",
            "matplotlib",
            "astropy",
            "astroquery",
            "photutils",
            "scikit-learn",
            "xarray",
            "tqdm",
            "h5py",
            "emcee",
            "corner",
            "pathos",
            "requests"
                    ],
    author="Hunter Brooks",
    author_email="hbrooks8@rockets.utoledo.edu",
    description="SphereX Data Retrieval and Photometry Toolkit",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/huntbrooks85/xcavation",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.10',
)




#-----------------------------------------------------------------------#
# xcavation.aperture v0.4.0
# By Hunter Brooks, at UToledo, Toledo: Jan. 28, 2026
#
# Purpose: Perform Aperture Photometry on SphereX Data
#-----------------------------------------------------------------------#



# Import Data Management
# ------------------------------------------------------ #
import numpy as np
import astropy.units as u
from astropy.io import fits
from scipy.interpolate import griddata
# ------------------------------------------------------ #



# Import Internal Modules
# ------------------------------------------------------ #
from .motion import time_mjd, proper_motion
# ------------------------------------------------------ #



# Import WCS, Photometry, and Plotting
# ------------------------------------------------------ #
from astropy.wcs import WCS
from lacosmic import remove_cosmics
from astropy.stats import SigmaClip
from astropy.coordinates import SkyCoord
from astropy.nddata.utils import Cutout2D
from scipy.ndimage import median_filter
from photutils.aperture import CircularAperture, CircularAnnulus, aperture_photometry, ApertureStats
# ------------------------------------------------------ #



# Import API Mangament Tools
# ------------------------------------------------------ #
import requests
from io import BytesIO
# ------------------------------------------------------ #




# Table of Resolving Powers Based on Wavelength
# ------------------------------------------------------ #
def resolving_table(wave):
  """
  Return the spectral resolution element (Δλ) for a given wavelength
  based on the SPHEREx instrument resolving powers. The resolving power,
  R = λ / Δλ, is taken from Akeson et al. (2025):
  https://irsa.ipac.caltech.edu/data/SPHEREx/docs/SPHEREx_Expsupp_QR_v1.0.pdf.

  Parameters
  ----------
    wave: Wavelength inputted to find corresponding R value (float)

  Returns
  -------
    delta_lambda: Spectral wavelength resolution at a given wavelength (float)
  """


  # ----- Table of R Values ----- #
  # Use Inputted Wavelength to Find Relevant R value
  if wave < 1.11:
    R = 39
  if 1.11 <= wave < 2.42:
    R = 41
  elif 2.42 <= wave < 3.82:
    R = 35
  elif 3.82 <= wave < 4.42:
    R = 110
  elif 4.42 <= wave:
    R = 130
  # ----------------------------- #



  # Calculate Delta Lambda Using Wavelength and R Value
  return wave/R
# ------------------------------------------------------ #




# Perform SphereX Wavelength and Aperature Calculation
# ------------------------------------------------------ #
def spherex_aperature_phot(url, coord, pm, mjd,  # Coords
                            r_fwhm, r_annulus_in, r_annulus_out, # Radii
                             ram_download, # Download Types
                              clean_type, bad_bits, # Bad Pixel Fixes
                                background_type): # Background Subraction
    """
    Perform SPHEREx aperture photometry for a given sky position and epoch.
    Downloads SPHEREx FITS file, propagates the input coordinates for
    proper motion, creates cutouts of flux, variance, and flags. Performs
    aperture photometry with a local background annulus. Spectral wavelength
    and resolution are also calulcated alongside counting bad pixels within
    the aperture.

    Parameters
    ----------
      url: URL of SPHEREx FITS file (str)
      coord : Input sky coordinates (RA, Dec) in degrees (list)
      pm: Proper motion in arcseconds per year (list)
      mjd: MJD corresponding to the input coordinates (float)
      r_fwhm: Aperture radius in units of PSF FWHM (float)
      r_annulus_in: Inner radius of background annulus in units of FWHM (float)
      r_annulus_out:  Outer radius of background annulus in units of FWHM (float)
      ram_download: Download the FITS entirely into memory (bool)
      bad_bits: FLAG extension to mask as bad pixels (list)

    Returns
    -------
      result : dict
          Dictionary containing the following keys:
          - 'wavelength': Wavelength at target position (float)
          - 'delta_lambda': Spectral resolution element Δλ (float)
          - 'flux': Aperture flux (float)
          - 'flux_err': Propagated flux uncertainty (float)
          - 'flag_count': Counts of flagged pixels per bad bit (dict)
          - 'SNR': Signal-to-noise ratio of the aperture flux (float)
          - 'flux_cutout': Flux cutout (array)
          - 'flag_cutout': FLAG cutout (array)
          - 'aperture': Binary aperture mask (bool array)
          - 'annulus': Binary annulus mask (bool array)
          - 'x_loc', 'y_loc': Pixel coordinates of target in cutout (float)
          - 'ap_radius': Aperture radius (float)
          - 'inner_annulus', 'outer_annulus': Annulus radii (float)
          - 'url': Input URL (str)
          - 'mjd': MJD of observation (float)
    """



    # ----- Download FITS entirely in RAM ----- #
    if ram_download is True:
      resp = requests.get(url)
      resp.raise_for_status()
      url = BytesIO(resp.content)
    # ----------------------------------------- #



    # ---------------- Open File ---------------- #
    with fits.open(url, lazy_load_hdus=True) as hdul:

        # ---------------- WCS ---------------- #
        # Gets Image Header and States Pixel Scale
        spectral_header = hdul["IMAGE"].header.copy()
        pixel_scale_deg=0.001708333 # Pixel Scale From Documentation

        # Blocks SIP warning message doesnt show
        for key in list(spectral_header.keys()):
            if key.startswith(("A_", "B_", "AP_", "BP_")):
                del spectral_header[key]

        # Gets Flag Data and Finds All Selected Bad Bits
        flags = hdul['FLAGS'].data # Flag Data
        bad_bitmask = sum(1 << bit for bit in bad_bits) # Bad Bits Data

        # Apply bad pixel mask and sets to 0
        bad_pixel_mask = (flags & bad_bitmask) != 0

        # Obtain Header Data
        celestial_wcs = WCS(hdul[1].header.copy()) # Normal WCS
        spectral_wcs = WCS(spectral_header, fobj=hdul, key="W") # Wavelength WCS

        # Obtain Flux Image
        flux_image_temp =  hdul[1].data.astype(float) # Flux Image
        zodi_image_temp = hdul[4].data.astype(float) # ZODE Image
        flux_image = flux_image_temp - zodi_image_temp # Image - ZODI
        flux_image = flux_image.astype(np.float32, copy=True)

        # Obtain Variance Image
        variance_image = hdul['VARIANCE'].data.astype(float) # VARIANCE Image
        variance_image = variance_image.astype(np.float32, copy=True)
        # ------------------------------------- #



        # ----- Proper Motion Propagation ----- #
        ra_in, dec_in = coord[0], coord[1] # Inputted R.A.
        pmra_in, pmdec_in = pm[0], pm[1] # Inputted Decl.
        mjd_obs = spectral_header['MJD-OBS'] # Observed MJD
        time_passed = time_mjd(mjd, mjd_obs) # Years since observation
        ra_deg, dec_deg = proper_motion(ra_in, dec_in, # Proper Motion Prop.
                                      pmra_in, pmdec_in, time_passed)
        coord = SkyCoord(ra_deg, dec_deg, unit='deg') # Convert to astropy Units
        # ------------------------------------- #



        # ---------------- Cutouts ---------------- #
        # Cutout of Flux, Variance and WCS Using Inputted Radius
        cutout_flux = Cutout2D(flux_image,
                                position=coord, size=35, wcs=celestial_wcs)

        # Returns Null Values if Cutout is Not Correct Size
        if cutout_flux.shape[0] != 35 or cutout_flux.shape[1] != 35:
          return {
                    "wavelength": np.nan,
                    "delta_lambda": np.nan,
                    "flux": np.nan,
                    "flux_err": np.nan,
                    "flag_count": np.nan,
                    'SNR': np.nan,
                    "flux_cutout": np.nan,
                    "flag_cutout": np.nan,
                    "aperture": np.nan,
                    "annulus": np.nan,
                    "x_loc": np.nan,
                    "y_loc": np.nan,
                    "ap_radius": np.nan,
                    "inner_annulus": np.nan,
                    "outer_annulus": np.nan,
                    "url": url,
                    "mjd": np.nan
                }

        # Cutout each Image, Flag, Variance, and WCS Frames
        cutout_var = Cutout2D(variance_image, # # Variance Cutout
                                position=coord, size=35, wcs=celestial_wcs)
        cutout_flag = Cutout2D(flags, # Flag Cutout
                                position=coord, size=35, wcs=celestial_wcs)
        cutout_mask = Cutout2D(bad_pixel_mask, # Selected Mask Cuout
                                position=coord, size=35, wcs=celestial_wcs)
        cutout_wcs = cutout_flux.wcs # WCS Cutout

        # Pull out just cutout data
        cutout_mask = cutout_mask.data
        cutout_flux = cutout_flux.data
        cutout_var = cutout_var.data
        # ----------------------------------------- #



        # ---------------- Interpolated Masking ---------------- #
        if clean_type == 'interp_mask':
          ny, nx = cutout_flux.shape
          y, x = np.mgrid[0:ny, 0:nx]

          # Good pixels only
          good = (~cutout_mask) & np.isfinite(cutout_flux)

          # Coordinates and values of good pixels
          points = np.column_stack((x[good], y[good]))
          values_flux = cutout_flux[good]
          values_var = cutout_var[good]

          # Interpolate over full grid
          interp_flux = griddata(points, values_flux, (x, y), method='nearest')
          interp_var = griddata(points, values_var, (x, y), method='nearest')
          cutout_flux[cutout_mask] = interp_flux[cutout_mask]
          cutout_var[cutout_mask] = interp_var[cutout_mask]
        # ------------------------------------------------------ #



        # ---------------- Median Masking ---------------- #
        if clean_type == 'median_mask':
          bad = cutout_flag != 0   # <-- THIS IS THE FIX
          med = median_filter(cutout_flux, size=5, mode='mirror')
          cutout_flux[bad] = med[bad]
        # ------------------------------------------------ #



        # ---------------- LaCosmic Masking ---------------- #
        # Clean Image and Variance Cutouts Using Astroscrappy
        if clean_type == 'lacosmic':
          from astropy import log
          log.setLevel('ERROR')

          cutout_flux, crmask = remove_cosmics(
                                              cutout_flux,                  # flux array
                                              contrast=1.5,                 # typical SPHEREx PSF sampling ~ 1–2 px
                                              cr_threshold=20,              # Laplacian S/N threshold
                                              neighbor_threshold=50,        # grows cosmic rays
                                              error=np.sqrt(cutout_var),    # variance array as 1-sigma error
                                              # mask=~cutout_mask,          # bad pixels / saturated stars
                                              maxiter=4,
                                              border_mode='nearest')
        del bad_pixel_mask # Delete Mask to Save on RAM
        # -------------------------------------------------- #




        # ---------------- Convert Units ---------------- #
        # Converts Units of Flux and Variance Cutout
        pixel_area_sr = (pixel_scale_deg * np.pi / 180)**2 # Pixel^2 to Sr
        flux_ujy = cutout_flux * 1e6 * pixel_area_sr * 1e6 # Flux Convert
        var_ujy = cutout_var * (1e6 * pixel_area_sr * 1e6)**2 # Variance Convert
        # ----------------------------------------------- #




        # ----- Set-Up Aperture Photometry ----- #
        # Gets x, y Pixel Locations and FWHM
        x_query, y_query = cutout_wcs.world_to_pixel(coord) # Cutouts Coords.
        fwhm = hdul[1].header['PSF_FWHM'] # Image FWHM

        # Aperature Set-Up
        ap_radius_arcsec = (r_fwhm * fwhm * u.arcsec).to(u.deg).value
        ap_radius_pix = ap_radius_arcsec / pixel_scale_deg # Aperture Radius
        aperture = CircularAperture((x_query, y_query), # Query Positon
                                    r=ap_radius_pix) # Radius

        # Annulus Set-up
        an_inner_arcsec = (r_annulus_in * fwhm * u.arcsec).to(u.deg).value
        an_outer_arcsec = (r_annulus_out * fwhm * u.arcsec).to(u.deg).value
        an_inner = an_inner_arcsec / pixel_scale_deg # Annulus Inner Radius
        an_outer = an_outer_arcsec / pixel_scale_deg # Annulus Outer Radius
        annulus_ap = CircularAnnulus((x_query, y_query), # Query Position
                                     r_in=an_inner, r_out=an_outer) # Radii
        # -------------------------------------- #



        # ----- Background Frame ----- #
        # Calculates Background Frame
        sigclip = SigmaClip(sigma=5.0, maxiters=5) # Sigma Clip For Background
        if background_type == 'mean':
          bkg_per_pix = (ApertureStats(flux_ujy, annulus_ap,
                          sigma_clip=sigclip)).mean # Average Annulus Background
        elif background_type == 'median':
          bkg_per_pix = (ApertureStats(flux_ujy, annulus_ap,
                          sigma_clip=sigclip)).median # Median Annulus Background
        elif background_type == 'mode':
          bkg_per_pix = (ApertureStats(flux_ujy, annulus_ap,
                          sigma_clip=sigclip)).mode # Mode Annulus Background

        ap_mask = aperture.to_mask(method='exact').to_image(flux_ujy.shape)

        # Calculate Aperture Area (w/ or w/o mask)
        if clean_type == 'mask':
          good_ap_mask = ap_mask * (~cutout_mask) # w/ mask
        else:
          good_ap_mask = ap_mask # w/o mask

        # Calcualtes Final Background Frame
        area = np.sum(good_ap_mask) # Total Aperture Area
        total_bkg = bkg_per_pix * area # Background Frame in Aperture Area
        # ---------------------------- #



        # ----- Final Aperture Photometry ----- #
        # Calculates the Aperature Photometry (w/ or w/o mask)
        if clean_type == 'mask':
          phot_table = aperture_photometry(flux_ujy, aperture,
                                           mask = cutout_mask) # w/ mask
        else:
          phot_table = aperture_photometry(flux_ujy, aperture) # w/o mask
        flux_ap = phot_table['aperture_sum'][0] - total_bkg # Final Aperture
        # ------------------------------------ #



        # ----- Aperture Error Photometry ----- #
        # Aperture Error (w/ or w/o mask)
        if clean_type == 'mask':
          phot_var_table = aperture_photometry(var_ujy, aperture,
                                                mask = cutout_mask) # w/ mask
        else:
          phot_var_table = aperture_photometry(var_ujy, aperture) # w/o mask

        # Calculate Progated Errors
        var_flux_ap = phot_var_table['aperture_sum'][0] # Aperture Error
        top_var = ((ApertureStats(flux_ujy, annulus_ap,
                                  sigma_clip=sigclip)).std)**2
        bot_var = (annulus_ap.area_overlap(var_ujy))
        var_bkg_per_pix = top_var/bot_var # Background Error
        flux_err = np.sqrt(var_flux_ap+(area**2)*var_bkg_per_pix) # Total Error
        # ------------------------------------- #



        # ----- Flag Aperture Count ----- #
        # Gets the Pixels in the Aperture
        ap_pixels = ap_mask > 0
        ap_flags = cutout_flag.data[ap_pixels]

        # Queried Array
        bit_counts = {} # Empty Array
        total_bad_bits = [0, 1, 2, 4, 5, 6, 7,
                          9, 10, 11, 12, 14, 15,
                          17, 19, 21]

        # Counts Each Flag in the Aperture
        for bit in total_bad_bits:
            bitmask = 1 << bit # Convert to bit
            bit_counts[bit] = np.count_nonzero(ap_flags & bitmask) # Count
        # ------------------------------- #



        # ----- Wavelength ----- #
        # General Set-Up
        y, x = np.indices(hdul["IMAGE"].data.shape) # Obtain Detector Size
        wave, *_ = spectral_wcs.pixel_to_world(x, y) # Wavelength WCS
        wave = np.asarray(wave) # Make Wavelength Arrary

        # Obtains Wavelength Average
        x_query_total, y_query_total = celestial_wcs.world_to_pixel(coord)
        wave_point = wave[int(y_query_total), int(x_query_total)] # Wavelength
        # ---------------------- #
    # ------------------------------------------- #



    # ----- Output Aperture/Annulus Array ----- #
    ap_mask_obj = aperture.to_mask(method='exact') # Aperture Mask
    an_mask_obj = annulus_ap.to_mask(method='exact') # Annulus Mask

    # Convert to image-sized arrays
    ap_mask = ap_mask_obj.to_image(cutout_flux.shape) # Aperture
    an_mask = an_mask_obj.to_image(cutout_flux.shape) # Annulus
    # ----------------------------------------- #



    # ----- Output Dictionary ----- #
    return {
        "wavelength": wave_point,
        "delta_lambda": resolving_table(wave_point),
        "flux": flux_ap,
        "flux_err": flux_err,
        "flag_count": bit_counts,
        "SNR": (flux_ap/flux_err),
        "flux_cutout": flux_ujy,
        "flag_cutout": cutout_flag.data,
        "aperture": ap_mask,
        "annulus": an_mask,
        "x_loc": x_query,
        "y_loc": y_query,
        "ap_radius": r_fwhm * fwhm,
        "inner_annulus": r_annulus_in * fwhm,
        "outer_annulus": r_annulus_out * fwhm,
        "url": url,
        "mjd": spectral_header['MJD-OBS']
    }
    # ----------------------------- #
# ------------------------------------------------------ #
